"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const chalk_1 = require("chalk");
const Debug = require("debug");
const cli_framework_1 = require("@ionic/cli-framework");
const cli_utils_1 = require("@ionic/cli-utils");
const command_1 = require("@ionic/cli-utils/lib/command");
const errors_1 = require("@ionic/cli-utils/lib/errors");
const debug = Debug('ionic:commands:link');
const CHOICE_CREATE_NEW_APP = 'createNewApp';
const CHOICE_NEVERMIND = 'nevermind';
const CHOICE_RELINK = 'relink';
const CHOICE_LINK_EXISTING_APP = 'linkExistingApp';
const CHOICE_IONIC = 'ionic';
const CHOICE_GITHUB = 'github';
const CHOICE_MASTER_ONLY = 'master';
const CHOICE_SPECIFIC_BRANCHES = 'specific';
const CHOICE_ALL_BRANCHES = 'all';
class LinkCommand extends command_1.Command {
    getMetadata() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return {
                name: 'link',
                type: 'project',
                description: 'Connect your local app to Ionic Pro',
                longDescription: `
If you have an app on Ionic Pro, you can link it to this local Ionic project with this command.

Excluding the ${chalk_1.default.green('app_id')} argument looks up your apps on Ionic Pro and prompts you to select one.

This command simply sets the ${chalk_1.default.bold('app_id')} property in ${chalk_1.default.bold(cli_utils_1.PROJECT_FILE)} for other commands to read.
      `,
                exampleCommands: ['', 'a1b2c3d4'],
                inputs: [
                    {
                        name: 'app_id',
                        description: `The ID of the app to link (e.g. ${chalk_1.default.green('a1b2c3d4')})`,
                    },
                ],
                options: [
                    {
                        name: 'name',
                        description: 'The app name to use during the linking of a new app',
                    },
                    {
                        name: 'create',
                        description: 'Create a new app on Ionic Pro and link it with this local Ionic project',
                        type: Boolean,
                        groups: [cli_utils_1.OptionGroup.Hidden],
                    },
                    {
                        name: 'pro-id',
                        description: 'Specify an app ID from the Ionic Pro to link',
                        groups: [cli_utils_1.OptionGroup.Hidden],
                    },
                ],
            };
        });
    }
    preRun(inputs, options) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { create } = options;
            if (inputs[0] && create) {
                throw new errors_1.FatalException(`Sorry--cannot use both ${chalk_1.default.green('app_id')} and ${chalk_1.default.green('--create')}. You must either link an existing app or create a new one.`);
            }
            const proAppId = options['pro-id'] || '';
            if (proAppId) {
                inputs[0] = proAppId;
            }
        });
    }
    run(inputs, options) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { promptToLogin } = yield Promise.resolve().then(() => require('@ionic/cli-utils/lib/session'));
            let [appId] = inputs;
            let { create } = options;
            const project = yield this.env.project.load();
            if (project.app_id) {
                if (project.app_id === appId) {
                    this.env.log.msg(`Already linked with app ${chalk_1.default.green(appId)}.`);
                    return;
                }
                const msg = appId ?
                    `Are you sure you want to link it to ${chalk_1.default.green(appId)} instead?` :
                    `Would you like to run link again?`;
                const confirm = yield this.env.prompt({
                    type: 'confirm',
                    name: 'confirm',
                    message: `App ID ${chalk_1.default.green(project.app_id)} is already set up with this app. ${msg}`,
                });
                if (!confirm) {
                    this.env.log.msg('Not linking.');
                    return;
                }
            }
            if (!(yield this.env.session.isLoggedIn())) {
                yield promptToLogin(this.env);
            }
            if (!appId && !create) {
                const choices = [
                    {
                        name: `Link ${project.app_id ? 'a different' : 'an existing'} app on Ionic Pro`,
                        value: CHOICE_LINK_EXISTING_APP,
                    },
                    {
                        name: 'Create a new app on Ionic Pro',
                        value: CHOICE_CREATE_NEW_APP,
                    },
                ];
                if (project.app_id) {
                    choices.unshift({
                        name: `Relink ${chalk_1.default.green(project.app_id)}`,
                        value: CHOICE_RELINK,
                    });
                }
                const result = yield this.env.prompt({
                    type: 'list',
                    name: 'whatToDo',
                    message: 'What would you like to do?',
                    choices,
                });
                if (result === CHOICE_CREATE_NEW_APP) {
                    create = true;
                }
                else if (result === CHOICE_LINK_EXISTING_APP) {
                    appId = yield this.chooseApp();
                    if (appId === CHOICE_NEVERMIND) {
                        this.env.log.msg('Not linking app.');
                    }
                }
                else if (result === CHOICE_RELINK) {
                    appId = project.app_id;
                }
            }
            if (create) {
                let name = options['name'] ? String(options['name']) : undefined;
                if (!name) {
                    name = yield this.env.prompt({
                        type: 'input',
                        name: 'name',
                        message: 'Please enter a name for your new app:',
                        validate: v => cli_framework_1.validators.required(v),
                    });
                }
                appId = yield this.createApp({ name });
            }
            else {
                const app = yield this.lookUpApp(appId);
                yield this.linkApp(app);
            }
            yield Promise.all([this.env.config.save(), this.env.project.save()]);
        });
    }
    getAppClient() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { AppClient } = yield Promise.resolve().then(() => require('@ionic/cli-utils/lib/app'));
            const token = yield this.env.session.getUserToken();
            return new AppClient({ token, client: this.env.client });
        });
    }
    getUserClient() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { UserClient } = yield Promise.resolve().then(() => require('@ionic/cli-utils/lib/user'));
            const token = yield this.env.session.getUserToken();
            return new UserClient({ token, client: this.env.client });
        });
    }
    lookUpApp(appId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.env.tasks.next(`Looking up app ${chalk_1.default.green(appId)}`);
            const appClient = yield this.getAppClient();
            const app = yield appClient.load(appId); // Make sure the user has access to the app
            this.env.tasks.end();
            return app;
        });
    }
    createApp({ name }) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const appClient = yield this.getAppClient();
            const app = yield appClient.create({ name });
            yield this.linkApp(app);
            return app.id;
        });
    }
    linkApp(app) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // TODO: load connections
            // TODO: check for git availability before this
            this.env.log.nl();
            this.env.log.info(`Ionic Pro uses a git-based workflow to manage app updates.\n` +
                `You will be prompted to set up the git host and repository for this new app. See the docs${chalk_1.default.cyan('[1]')} for more information.\n\n` +
                `${chalk_1.default.cyan('[1]')}: ${chalk_1.default.bold('https://ionicframework.com/docs/pro/basics/git/')}`);
            const service = yield this.env.prompt({
                type: 'list',
                name: 'gitService',
                message: 'Which git host would you like to use?',
                choices: [
                    {
                        name: 'Github',
                        value: CHOICE_GITHUB,
                    },
                    {
                        name: 'Ionic Pro',
                        value: CHOICE_IONIC,
                    },
                ],
            });
            if (service === CHOICE_IONIC) {
                const config = yield this.env.config.load();
                if (!config.git.setup) {
                    yield this.env.runCommand(['ssh', 'setup']);
                }
                yield this.env.runCommand(['config', 'set', 'app_id', `"${app.id}"`, '--json']);
                yield this.env.runCommand(['git', 'remote']);
            }
            else {
                if (service === CHOICE_GITHUB) {
                    yield this.linkGithub(app);
                }
                yield this.env.runCommand(['config', 'set', 'app_id', `"${app.id}"`, '--json']);
            }
            this.env.log.ok(`Project linked with app ${chalk_1.default.green(app.id)}!`);
        });
    }
    linkGithub(app) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { id } = yield this.env.session.getUser();
            const userClient = yield this.getUserClient();
            const user = yield userClient.load(id, { fields: ['oauths'] });
            if (!user.oauths || !user.oauths.includes('github')) {
                yield this.oAuthProcess(id);
            }
            if (yield this.needsAssociation(app, user.id)) {
                yield this.confirmGithubRepoExists();
                const repoId = yield this.selectGithubRepo();
                const branches = yield this.selectGithubBranches(repoId);
                yield this.connectGithub(app, repoId, branches);
            }
        });
    }
    confirmGithubRepoExists() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let confirm = false;
            this.env.log.nl();
            this.env.log.info(chalk_1.default.bold(`In order to link to a GitHub repository the repository must already exist on GitHub.`));
            this.env.log.info(`${chalk_1.default.bold('If the repository does not exist please create one now before continuing.')}\n` +
                `If you're not familiar with Git you can learn how to set it up with GitHub here:\n\n` +
                chalk_1.default.cyan(`https://help.github.com/articles/set-up-git/ \n\n`) +
                `You can find documenation on how to create a repository on GitHub and push to it here:\n\n` +
                chalk_1.default.cyan(`https://help.github.com/articles/create-a-repo/`));
            confirm = yield this.env.prompt({
                type: 'confirm',
                name: 'confirm',
                message: 'Does the repository exist on GitHub?',
            });
            if (!confirm) {
                throw new errors_1.FatalException('Aborting.');
            }
        });
    }
    oAuthProcess(userId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const opn = yield Promise.resolve().then(() => require('opn'));
            const userClient = yield this.getUserClient();
            let confirm = false;
            this.env.log.nl();
            this.env.log.info(`Github OAuth setup required.\n` +
                `To continue, we need you to authorize Ionic Pro with your Github account. ` +
                `A browser will open and prompt you to complete the authorization request. ` +
                `When finished, please return to the CLI to continue linking your app.`);
            confirm = yield this.env.prompt({
                type: 'confirm',
                name: 'ready',
                message: 'Open browser:',
            });
            if (!confirm) {
                throw new errors_1.FatalException('Aborting.');
            }
            const url = yield userClient.oAuthGithubLogin(userId);
            opn(url, { wait: false });
            confirm = yield this.env.prompt({
                type: 'confirm',
                name: 'ready',
                message: 'Authorized and ready to continue:',
            });
            if (!confirm) {
                throw new errors_1.FatalException('Aborting.');
            }
        });
    }
    needsAssociation(app, userId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const appClient = yield this.getAppClient();
            if (app.association && app.association.repository.html_url) {
                this.env.log.msg(`App ${chalk_1.default.green(app.id)} already connected to ${chalk_1.default.bold(app.association.repository.html_url)}`);
                const confirm = yield this.env.prompt({
                    type: 'confirm',
                    name: 'confirm',
                    message: 'Would you like to connect a different repo?',
                });
                if (!confirm) {
                    return false;
                }
                try {
                    // TODO: maybe we can use a PUT instead of DELETE now + POST later?
                    yield appClient.deleteAssociation(app.id);
                }
                catch (e) {
                    if (cli_utils_1.isSuperAgentError(e)) {
                        if (e.response.status === 401) {
                            yield this.oAuthProcess(userId);
                            yield appClient.deleteAssociation(app.id);
                            return true;
                        }
                        else if (e.response.status === 404) {
                            debug(`DELETE ${app.id} Github association not found`);
                            return true;
                        }
                    }
                    throw e;
                }
            }
            return true;
        });
    }
    connectGithub(app, repoId, branches) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const appClient = yield this.getAppClient();
            try {
                const association = yield appClient.createAssociation(app.id, { repoId, type: 'github', branches });
                this.env.log.ok(`App ${chalk_1.default.green(app.id)} connected to ${chalk_1.default.bold(association.repository.html_url)}`);
            }
            catch (e) {
                if (cli_utils_1.isSuperAgentError(e) && e.response.status === 403) {
                    throw new errors_1.FatalException(e.response.body.error.message);
                }
            }
        });
    }
    formatRepoName(fullName) {
        const [org, name] = fullName.split('/');
        return `${chalk_1.default.dim(`${org} /`)} ${name}`;
    }
    chooseApp() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { formatName } = yield Promise.resolve().then(() => require('@ionic/cli-utils/lib/app'));
            this.env.tasks.next(`Looking up your apps`);
            const apps = [];
            const appClient = yield this.getAppClient();
            const paginator = appClient.paginate();
            for (const r of paginator) {
                const res = yield r;
                apps.push(...res.data);
            }
            this.env.tasks.end();
            const neverMindChoice = {
                name: 'Nevermind',
                id: CHOICE_NEVERMIND,
                org: null,
            };
            const linkedApp = yield this.env.prompt({
                type: 'list',
                name: 'linkedApp',
                message: 'Which app would you like to link',
                choices: [neverMindChoice, ...apps].map(app => ({
                    name: CHOICE_NEVERMIND === app.id ? chalk_1.default.bold(app.name) : `${formatName(app)} ${chalk_1.default.dim(`(${app.id})`)}`,
                    value: app.id,
                })),
            });
            return linkedApp;
        });
    }
    selectGithubRepo() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const user = yield this.env.session.getUser();
            const userClient = yield this.getUserClient();
            this.env.tasks.next('Looking up your Github repositories');
            const paginator = userClient.paginateGithubRepositories(user.id);
            const repos = [];
            try {
                for (const r of paginator) {
                    const res = yield r;
                    repos.push(...res.data);
                    this.env.tasks.updateMsg(`Looking up your Github repositories: ${chalk_1.default.bold(String(repos.length))} found`);
                }
            }
            catch (e) {
                this.env.tasks.fail();
                if (cli_utils_1.isSuperAgentError(e) && e.response.status === 401) {
                    yield this.oAuthProcess(user.id);
                    return this.selectGithubRepo();
                }
                throw e;
            }
            this.env.tasks.end();
            const repoId = yield this.env.prompt({
                type: 'list',
                name: 'githubRepo',
                message: 'Which Github repository would you like to link?',
                choices: repos.map(repo => ({
                    name: this.formatRepoName(repo.full_name),
                    value: String(repo.id),
                })),
            });
            return Number(repoId);
        });
    }
    selectGithubBranches(repoId, linkedBranches = [], availableBranches = []) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // If the array is empty this is the first time this was called so we should explain the process
            // else see if they need to add more branches
            if (linkedBranches.length === 0) {
                this.env.log.nl();
                this.env.log.info(chalk_1.default.bold(`By default Ionic Pro links only to the ${chalk_1.default.green('master')} branch.`));
                this.env.log.info(`${chalk_1.default.bold('If you\'d like to link to another branch or multiple branches you\'ll need to select each branch to connect to.')}\n` +
                    `If you're not familiar with on working with branches in GitHub you can read about them here:\n\n` +
                    chalk_1.default.cyan(`https://guides.github.com/introduction/flow/ \n\n`));
                const choice = yield this.env.prompt({
                    type: 'list',
                    name: 'githubMultipleBranches',
                    message: 'Which would you like to do?',
                    choices: [
                        {
                            name: `Link to ${chalk_1.default.green('master')} branch only`,
                            value: CHOICE_MASTER_ONLY,
                        },
                        {
                            name: `Link to specific branches`,
                            value: CHOICE_SPECIFIC_BRANCHES,
                        },
                        {
                            name: `Link to all branches`,
                            value: CHOICE_ALL_BRANCHES,
                        },
                    ],
                });
                switch (choice) {
                    case CHOICE_MASTER_ONLY:
                        return ['master'];
                    case CHOICE_SPECIFIC_BRANCHES:
                        // fall through and begin prompting to choose branches
                        break;
                    case CHOICE_ALL_BRANCHES:
                        return ['*'];
                    default:
                        throw new errors_1.FatalException('Aborting. No branch choice specified.');
                }
            }
            else {
                const confirm = yield this.env.prompt({
                    type: 'confirm',
                    name: 'confirm',
                    message: `Would you like to link additional branches?`,
                });
                if (!confirm) {
                    return linkedBranches;
                }
            }
            // If no available branches were passed in look for some
            if (availableBranches.length === 0) {
                const user = yield this.env.session.getUser();
                const userClient = yield this.getUserClient();
                const paginator = userClient.paginateGithubBranches(user.id, repoId);
                this.env.tasks.next('Looking for available branches');
                try {
                    for (const r of paginator) {
                        const res = yield r;
                        availableBranches.push(...res.data);
                        this.env.tasks.updateMsg(`Looking up the available branches on your Github repository: ${chalk_1.default.bold(String(availableBranches.length))} found`);
                    }
                }
                catch (e) {
                    this.env.tasks.fail();
                    throw e;
                }
                this.env.tasks.end();
            }
            // filter out already chosen branches from available list
            availableBranches = availableBranches.filter(branch => !linkedBranches.includes(branch.name));
            const choices = availableBranches.map(branch => ({
                name: branch.name,
                value: branch.name,
            }));
            // NOTE: the ~ is not a valid branch char so this will never conflict with existing branches
            const CHOICE_DONE_LINKING = '~done';
            // let them back out if they change there mind and have selected at least 1 branch
            if (linkedBranches.length > 0) {
                choices.push({ name: 'Done linking branches', value: CHOICE_DONE_LINKING });
            }
            const selectedBranch = yield this.env.prompt({
                type: 'list',
                name: 'githubBranches',
                message: 'Which Github repository would you like to link?',
                choices: choices,
            });
            if (selectedBranch === CHOICE_DONE_LINKING) {
                return linkedBranches;
            }
            linkedBranches.push(selectedBranch);
            return this.selectGithubBranches(repoId, linkedBranches, availableBranches);
        });
    }
}
exports.LinkCommand = LinkCommand;
